# Authors: Florian Hofer
#
# License: BSD (3-clause)

import hashlib
from pathlib import Path
from typing import Optional

import requests

__all__ = [
    'download_file',
]

_HASH_FUNCTIONS = {
    'md5': hashlib.md5,
    'sha256': hashlib.sha256,
}


def _verify_checksum(
    filepath: Path,
    checksum: str,
    checksum_type: str,
) -> bool:
    """
    Verify a file with a given checksum.

    Parameters
    ----------
    filepath : Path
        Path to the file to verify.
    checksum : str
        The checksum, against which the file will be verified.
    checksum_type : {'md5', 'sha256'}
        Type of the checksum.

    Returns
    -------
    bool
        `False` for nonexisting file or failed verification, `True`
        otherwise.
    """
    if not filepath.is_file():
        return False

    computed_hash = _HASH_FUNCTIONS[checksum_type]()
    with open(filepath, 'rb') as file:
        while True:
            chunk = file.read(8192)
            if not chunk:
                break
            computed_hash.update(chunk)
    return checksum == computed_hash.hexdigest()


def download_file(
    url: str,
    target_filepath: Path,
    checksum: Optional[str] = None,
    checksum_type: Optional[str] = None,
    verbose: bool = False,
):
    """
    Download a single file from `url` to `target_filepath`.

    In case `checksum` and `checksum_type` are provided, the downloaded
    file is verified.

    Parameters
    ----------
    url : str
        URL to download from.
    target_filepath : Path
        Location where the downloaded file will be stored.
    checksum : Optional[str], optional
        Checksum to verify the file against, by default None.
    checksum_type : Optional[str], optional
        Type of the checksum, by default None.
    verbose : bool, optional
        If `True`, output information during download. By default False.
    """
    if checksum is not None and checksum_type is not None:
        if _verify_checksum(target_filepath, checksum, checksum_type):
            if verbose:
                print(f'Skipping {url}, already downloaded')
            return

    target_filepath.parent.mkdir(parents=True, exist_ok=True)

    if verbose:
        print(f'Downloading {url}...')

    response = requests.get(url)
    response.raise_for_status()

    with open(target_filepath, 'wb') as file:
        file.write(response.content)

    if checksum is not None and checksum_type is not None:
        if not _verify_checksum(target_filepath, checksum, checksum_type):
            print(f'checksum verification failed for {target_filepath}, downloading again')
            download_file(url, target_filepath, checksum, checksum_type)
